import time
from typing import Any, Dict, Generic, List, Optional, TypeVar, Union

from pydantic import BaseModel, Field, model_validator
from shortuuid import uuid

T = TypeVar("T", bound=BaseModel)

# === Base ===


class BaseRequest(BaseModel):
    """Base request"""

    type: str
    request_id: Optional[str] = None
    output_topic: Optional[str] = None
    output_partition: Optional[int] = None


class BaseResponse(BaseModel):
    """Base response"""

    type: str
    request_id: str


# === Chat ===


class ImageUrlContent(BaseModel):
    """Image URL content for chat requests"""

    url: str


class ContentItem(BaseModel):
    """Content item for chat requests"""

    type: str
    text: Optional[str] = None
    image_url: Optional[ImageUrlContent] = None


class MessageItem(BaseModel):
    """Message item for chat requests"""

    role: str
    content: Union[str, List[ContentItem]]  # Updated to allow a list of ContentItem


class Prompt(BaseModel):
    """Prompt for chat requests"""

    messages: List[MessageItem]


class SamplingParams(BaseModel):
    """Sampling parameters for chat requests"""

    n: int = 1
    best_of: Optional[int] = None
    presence_penalty: float = 0.0
    frequency_penalty: float = 0.0
    repetition_penalty: float = 1.0
    temperature: float = 1.0
    top_p: float = 1.0
    top_k: int = -1
    min_p: float = 0.0
    seed: Optional[int] = None
    stop: Optional[List[str]] = None
    stop_token_ids: Optional[List[int]] = None
    min_tokens: int = 0
    logprobs: Optional[int] = None
    prompt_logprobs: Optional[int] = None
    detokenize: bool = True
    skip_special_tokens: bool = True
    spaces_between_special_tokens: bool = True
    truncate_prompt_tokens: Optional[int] = None


class Usage(BaseModel):
    """Usage for chat requests"""

    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class ChatRequest(BaseRequest):
    """Chat request"""

    type: str = "ChatRequest"
    model: Optional[str] = None
    kind: Optional[str] = None
    provider: Optional[str] = None
    namespace: Optional[str] = None
    adapter: Optional[str] = None
    prompt: Optional[Prompt] = None
    batch: Optional[List[Prompt]] = None
    max_tokens: int = Field(default=512)
    sampling_params: SamplingParams = Field(default_factory=SamplingParams)
    stream: bool = False
    user_id: Optional[str] = None
    organizations: Optional[Dict[str, Dict[str, str]]] = None
    handle: Optional[str] = None

    @model_validator(mode="before")
    @classmethod
    def replace_none_with_default(cls, values: dict) -> dict:
        if "max_tokens" in values and values["max_tokens"] is None:
            values["max_tokens"] = 512
        if "sampling_params" in values and values["sampling_params"] is None:
            values["sampling_params"] = SamplingParams()
        return values


class Choice(BaseModel):
    """Individual choice in the token response"""

    index: int
    text: str
    tokens: Optional[List[str]] = None
    token_ids: Optional[List[int]] = None
    logprobs: Optional[List[Dict[Union[int, str], Any]]] = None
    finish_reason: Optional[str] = None


class ChatResponse(BaseResponse, Generic[T]):
    """Chat response"""

    type: str = "ChatResponse"
    choices: List[Choice]
    trip_time: Optional[float] = None
    usage: Optional[Usage] = None
    parsed: Optional[T] = None


class TokenResponse(BaseResponse):
    """Token response"""

    type: str = "TokenResponse"
    choices: List[Choice]
    usage: Optional[Usage] = None


class V1ChatEvent(BaseModel):
    """
    Chat event containing both the request and response, along with additional metadata.
    """

    type: str = "ChatEvent"
    id: str = Field(default_factory=lambda: str(uuid()))
    request: ChatRequest
    response: ChatResponse
    token_count: Optional[int] = None
    trip_time: Optional[float] = None
    approved: Optional[bool] = None
    metadata: Optional[Dict[str, str]] = None
    owner_id: Optional[str] = None
    organization_id: Optional[str] = None
    handle: Optional[str] = None
    created: Optional[int] = None


# === Completion ===


class CompletionRequest(BaseModel):
    """Request for completion requests"""

    text: str
    images: Optional[List[str]] = None


class CompletionResponse(BaseResponse):
    """Completion response"""

    type: str = "CompletionResponse"
    choices: List[Choice]
    trip_time: Optional[float] = None
    usage: Optional[Usage] = None


# === OCR ===


class OCRRequest(BaseRequest):
    """Simple OCR request following EasyOCR patterns"""

    type: str = "OCRRequest"
    model: Optional[str] = None
    provider: Optional[str] = None
    image: str
    languages: List[str]  # e.g. ['en'], ['ch_sim', 'en']
    gpu: bool = True
    detail: bool = True  # True returns bounding boxes, False returns just text
    paragraph: bool = False  # Merge text into paragraphs
    min_confidence: Optional[float] = 0.0


class BoundingBox(BaseModel):
    """Coordinates for text location: [[x1,y1], [x2,y2], [x3,y3], [x4,y4]]"""

    points: List[List[int]]  # List of 4 points (8 coordinates total)
    text: str
    confidence: float


class OCRResponse(BaseResponse):
    """Response containing detected text and locations"""

    type: str = "OCRResponse"
    results: Union[List[BoundingBox], List[str]]  # List[str] if detail=False
    processing_time: Optional[float] = None
    usage: Optional[Usage] = None


# === Embeddings ===


class EmbeddingRequest(BaseRequest):
    """Embedding request"""

    type: str = "EmbeddingRequest"
    model: Optional[str] = None
    provider: Optional[str] = None
    text: Optional[str] = None
    image: Optional[str] = None


class Embedding(BaseModel):
    """Embedding"""

    object: str
    index: int
    embedding: List[float]


class EmbeddingResponse(BaseResponse):
    """Embedding response"""

    type: str = "EmbeddingResponse"
    object: str
    data: List[Embedding]
    model: str
    usage: Optional[Usage] = None


# === Deployments ===


class V1ReplicasInfo(BaseModel):
    """
    Replica status information for a given deployment.
    """

    desired: int = 0
    ready: int = 0
    updated: int = 0
    available: int = 0
    unavailable: int = 0


class V1VLLMParams(BaseModel):
    """
    Configuration parameters for vLLM-based deployments.
    """

    model: str
    model_type: Optional[str] = None
    trust_remote_code: bool = True
    tensor_parallel_size: int = 1
    dtype: str = "auto"
    max_images_per_prompt: int = 1
    device: str = "cuda"
    max_model_len: int = 8192
    max_num_seqs: int = 5
    gpu_memory_utilization: float = 0.8
    enforce_eager: bool = True
    enable_adapter: bool = False


class V1SentenceTFParams(BaseModel):
    """
    Configuration parameters for SentenceTransformers-based deployments.
    """

    model: str = "clip-ViT-B-32"
    device: str = "cuda"


class V1DoctrParams(BaseModel):
    """
    Configuration parameters for doctr-based OCR deployments.
    """

    det_arch: str = "fast_base"
    reco_arch: str = "crnn_vgg16_bn"
    pretrained: bool = True


class V1EasyOCRParams(BaseModel):
    """
    Configuration parameters for EasyOCR-based deployments.
    """

    device: str = "cuda"
    gpu: bool = True
    lang_list: List[str] = Field(default_factory=lambda: ["en"])
    quantize: bool = False


class V1LiteLLMParams(BaseModel):
    """
    Configuration parameters for LiteLLM-based deployments.
    """

    api_keys: Dict[str, str] = Field(default_factory=dict)


class V1ModelDeployment(BaseModel):
    """
    A model deployment, containing configuration and current status for a running model.
    """

    id: str = ""
    namespace: str = ""
    provider: str = ""
    kind: str = ""
    replicas: Optional[V1ReplicasInfo] = None
    status: str = ""
    image: str = ""
    vram_request: str = ""
    gpu_type: Optional[str] = None
    cpu_request: Optional[str] = None
    vllm_params: Optional[V1VLLMParams] = None
    sentence_tf_params: Optional[V1SentenceTFParams] = None
    doctr_params: Optional[V1DoctrParams] = None
    easyocr_params: Optional[V1EasyOCRParams] = None
    litellm_params: Optional[V1LiteLLMParams] = None


class V1ModelDeploymentRequest(BaseModel):
    """
    Request structure for creating or updating a model deployment.
    """

    namespace: Optional[str] = None
    provider: str
    vram_request: str
    memory_request: Optional[str] = None
    cpu_request: Optional[str] = None
    vllm_params: Optional[V1VLLMParams] = None
    sentence_tf_params: Optional[V1SentenceTFParams] = None
    doctr_params: Optional[V1DoctrParams] = None
    easyocr_params: Optional[V1EasyOCRParams] = None
    litellm_params: Optional[V1LiteLLMParams] = None
    max_pixels: Optional[int] = None


# === Training & Buffers ===


class V1LlamaFactoryParams(BaseModel):
    """
    LlamaFactoryParams
    """

    model: str = ""


class V1MSSwiftParams(BaseModel):
    """
    MSSwiftParams
    """

    model: str = "Qwen/Qwen2-VL-7B-Instruct"
    model_type: str = "qwen2-vl-7b-instruct"
    train_type: str = "lora"
    deepspeed: str = "zero3"
    torch_dtype: str = "bfloat16"
    max_length: int = 8192
    dataset: str = ""
    val_split_ratio: float = 0.90
    num_train_epochs: int = 3
    eval_strategy: str = "epoch"
    save_strategy: str = "epoch"
    save_total_limit: int = 3
    lora_rank: Optional[int] = None
    lora_alpha: Optional[int] = None
    size_factor: int = 28
    max_pixels: int = 1025000
    resume_from_checkpoint: Optional[str] = None
    freeze_vit: Optional[bool] = None
    rlhf_type: Optional[str] = None


class V1TrainingRequest(BaseModel):
    """
    Request to create a new training job
    """

    name: Optional[str] = None
    namespace: Optional[str] = None
    provider: str
    vram_request: str
    cpu_request: Optional[str] = None
    trust_remote_code: Optional[bool] = None
    adapter: Optional[str] = None
    buffer: Optional[str] = None
    queue: Optional[str] = None
    ms_swift_params: Optional[V1MSSwiftParams] = None
    llama_factory_params: Optional[V1LlamaFactoryParams] = None
    resume: Optional[bool] = None
    labels: Optional[Dict[str, str]] = None


class V1Checkpoint(BaseModel):
    """
    Checkpoint with step count and optional artifacts
    """

    step: int = 0
    trainer_state: Optional[str] = None
    args: Optional[str] = None
    adapter_config: Optional[str] = None


class V1TrainingJob(BaseModel):
    """
    Information about an existing training job
    """

    id: str
    name: str = ""
    namespace: str = ""
    provider: str = ""
    vram_request: str = ""
    cpu_request: Optional[str] = None
    trust_remote_code: Optional[bool] = None
    adapter: Optional[str] = None
    buffer: Optional[str] = None
    ms_swift_params: Optional[V1MSSwiftParams] = None
    llama_factory_params: Optional[V1LlamaFactoryParams] = None
    resume: Optional[bool] = None
    status: str = ""
    checkpoints: Optional[List[V1Checkpoint]] = None
    queue: Optional[str] = None
    labels: Optional[Dict[str, str]] = None
    created: int = Field(default_factory=lambda: int(time.time()))
    updated: int = Field(default_factory=lambda: int(time.time()))


class V1TrainingJobsResponse(BaseModel):
    """
    A list of training jobs
    """

    jobs: List[V1TrainingJob] = Field(default_factory=list)


class V1MSSwiftBufferParams(BaseModel):
    """
    MSSwiftParams variant for buffer-based training
    """

    model: str = "Qwen/Qwen2-VL-7B-Instruct"
    model_type: str = "qwen2-vl-7b-instruct"
    train_type: str = "lora"
    deepspeed: str = "zero3"
    torch_dtype: str = "bfloat16"
    max_length: int = 8192
    val_split_ratio: float = 0.90
    num_train_epochs: int = 3
    eval_strategy: str = "epoch"
    save_strategy: str = "epoch"
    save_total_limit: int = 3
    lora_rank: Optional[int] = None
    lora_alpha: Optional[int] = None
    size_factor: int = 28
    max_pixels: int = 1025000
    resume_from_checkpoint: Optional[str] = None
    freeze_vit: Optional[bool] = None
    rlhf_type: Optional[str] = None


class V1ReplayBufferRequest(BaseModel):
    """
    Request to create a replay buffer for incremental training
    """

    name: str
    namespace: Optional[str] = None
    provider: str
    vram_request: str
    cpu_request: Optional[str] = None
    trust_remote_code: Optional[bool] = None
    adapter: Optional[str] = None
    train_every: int = 50
    sample_n: int = 100
    sample_strategy: str = "Random"
    queue: Optional[str] = None
    ms_swift_params: Optional[V1MSSwiftBufferParams] = None
    llama_factory_params: Optional[V1LlamaFactoryParams] = None
    labels: Optional[Dict[str, str]] = None


class V1UpdateReplayBufferRequest(BaseModel):
    """
    Request to update a replay buffer for incremental training
    """

    name: Optional[str] = None
    namespace: Optional[str] = None
    provider: Optional[str] = None
    vram_request: Optional[str] = None
    cpu_request: Optional[str] = None
    trust_remote_code: Optional[bool] = None
    adapter: Optional[str] = None
    train_every: Optional[int] = None
    sample_n: Optional[int] = None
    sample_strategy: Optional[str] = None
    queue: Optional[str] = None
    ms_swift_params: Optional[V1MSSwiftBufferParams] = None
    llama_factory_params: Optional[V1LlamaFactoryParams] = None
    labels: Optional[Dict[str, str]] = None


class V1ReplayBuffer(BaseModel):
    """
    A replay buffer used for incremental training
    """

    id: str = ""
    name: str = ""
    namespace: str = ""
    provider: str = ""
    vram_request: str = ""
    cpu_request: Optional[str] = None
    trust_remote_code: Optional[bool] = None
    adapter: Optional[str] = None
    train_every: Optional[int] = None
    sample_n: int = 100
    sample_strategy: str = "Random"
    ms_swift_params: Optional[V1MSSwiftBufferParams] = None
    llama_factory_params: Optional[V1LlamaFactoryParams] = None
    labels: Optional[Dict[str, str]] = None
    queue: Optional[str] = None
    num_records: Optional[int] = None
    train_idx: Optional[int] = None


class V1ReplayBuffersResponse(BaseModel):
    """
    A list of replay buffers
    """

    buffers: List[V1ReplayBuffer] = Field(default_factory=list)


class V1ReplayBufferData(BaseModel):
    """
    The data stored in the replay buffer
    """

    examples: List[dict] = Field(default_factory=list)
    train: Optional[bool] = None


class V1Adapters(BaseModel):
    """
    A list of known adapters
    """

    adapters: List[str] = Field(default_factory=list)


class V1Models(BaseModel):
    """
    A list of known models
    """

    models: List[str] = Field(default_factory=list)


class V1Datasets(BaseModel):
    """
    A list of known datasets
    """

    datasets: List[str] = Field(default_factory=list)


class V1ModelFile(BaseModel):
    """
    Information about the latest checkpoint file
    """

    latest_checkpoint: Optional[str] = None


# === Errors ===


class ErrorResponse(BaseResponse):
    """Error response"""

    type: str = "ErrorResponse"
    error: str
    traceback: Optional[str] = None


# === Model ===


class ModelReadyResponse(BaseResponse):
    """Response indicating if a model is ready"""

    type: str = "ModelReadyResponse"
    request_id: str
    ready: bool
    error: Optional[str] = None



from __future__ import annotations

import os
from typing import Optional

import yaml


class Config:
    ORIGN_ADDR = os.getenv("ORIGN_ADDR", "https://orign.agentlabs.xyz")
    AGENTSEA_API_KEY = os.getenv("AGENTSEA_API_KEY")


class GlobalConfig:
    def __init__(
        self,
        api_key: Optional[str] = None,
        server: Optional[str] = None,
        debug: bool = False,
    ):
        self.api_key = api_key or Config.AGENTSEA_API_KEY
        self.server = server or Config.ORIGN_ADDR
        self.debug = debug

    def write(self) -> None:
        home = os.path.expanduser("~")
        dir = os.path.join(home, ".agentsea")
        os.makedirs(dir, exist_ok=True)
        path = os.path.join(dir, "orign.yaml")

        with open(path, "w") as yaml_file:
            yaml.dump(self.__dict__, yaml_file)
            yaml_file.flush()
            yaml_file.close()

    @classmethod
    def read(cls) -> GlobalConfig:
        home = os.path.expanduser("~")
        dir = os.path.join(home, ".agentsea")
        os.makedirs(dir, exist_ok=True)
        path = os.path.join(dir, "orign.yaml")

        if not os.path.exists(path):
            return GlobalConfig()

        with open(path, "r") as yaml_file:
            config = yaml.safe_load(yaml_file)
            return GlobalConfig(**config)



import base64
import json
from io import BytesIO
from typing import List, Optional
from urllib.parse import quote

import websockets
from PIL import Image

from orign.config import GlobalConfig
from orign.models import ErrorResponse, ModelReadyResponse, OCRRequest, OCRResponse


class AsyncOCRModel:
    def __init__(
        self,
        model: Optional[str] = None,
        provider: Optional[str] = None,
    ):
        config = GlobalConfig.read()
        self.model = model
        self.provider = provider
        self.orign_host = config.server.strip("http://").strip("https://")
        self.websocket = None

        # Construct the WebSocket URL
        self.ws_url = f"wss://{self.orign_host}/v1/ocr/stream"
        params = []
        if self.model:
            params.append(f"model={quote(self.model)}")
        if self.provider:
            params.append(f"provider={quote(self.provider)}")

        if params:
            self.ws_url += "?" + "&".join(params)

    async def connect(self):
        """Establish WebSocket connection if not already connected"""
        if not self.websocket:
            config = GlobalConfig.read()
            headers = {"Authorization": f"Bearer {config.api_key}"}
            self.websocket = await websockets.connect(
                self.ws_url, extra_headers=headers
            )

    async def close(self):
        """Close the WebSocket connection"""
        if self.websocket:
            await self.websocket.close()
            self.websocket = None

    async def detect(
        self,
        image: str,  # can be a URL or a base64 encoded image
        languages: List[str] = ["en"],  # e.g. ['en'], ['ch_sim', 'en']
        gpu: bool = True,
        detail: bool = True,  # True returns bounding boxes, False returns just text
        paragraph: bool = False,  # Merge text into paragraphs
        min_confidence: Optional[float] = 0.0,
    ) -> OCRResponse:
        await self.connect()

        if isinstance(image, Image.Image):
            buffered = BytesIO()
            image.save(buffered, format="PNG")
            image = "data:image/png;base64," + base64.b64encode(
                buffered.getvalue()
            ).decode("utf-8")

        request = OCRRequest(
            image=image,
            languages=languages,
            gpu=gpu,
            detail=detail,
            paragraph=paragraph,
            min_confidence=min_confidence,
        )

        if not self.websocket:
            raise ConnectionError("WebSocket connection not established")

        # Send the request
        await self.websocket.send(request.model_dump_json())

        # Yield responses as they arrive
        try:
            while True:
                response = await self.websocket.recv()
                response_data = json.loads(response)

                if response_data.get("type") == OCRResponse.__name__:
                    return OCRResponse.model_validate(response_data)

                elif response_data.get("type") == ModelReadyResponse.__name__:
                    ready = ModelReadyResponse.model_validate(response_data)
                    if not ready.ready:
                        raise ValueError(f"Model not ready: {ready.error}")

                elif response_data.get("type") == ErrorResponse.__name__:
                    error = ErrorResponse.model_validate(response_data)
                    raise ValueError(f"Error: {error.error}")
                else:
                    raise ValueError(f"Unknown response type: {response_data}")

        except websockets.exceptions.ConnectionClosed:
            await self.close()
            raise ConnectionError("WebSocket connection closed unexpectedly")



import base64
import json
from io import BytesIO
from typing import List, Optional
from urllib.parse import quote

import websocket
from PIL import Image

from orign.config import GlobalConfig
from orign.models import ErrorResponse, ModelReadyResponse, OCRRequest, OCRResponse


class OCRModel:
    def __init__(
        self,
        model: Optional[str] = None,
        provider: Optional[str] = None,
    ):
        config = GlobalConfig.read()
        self.model = model
        self.provider = provider
        self.orign_host = config.server.strip("http://").strip("https://")
        self.websocket = None

        # Construct the WebSocket URL
        self.ws_url = f"wss://{self.orign_host}/v1/ocr/stream"
        params = []
        if self.model:
            params.append(f"model={quote(self.model)}")
        if self.provider:
            params.append(f"provider={quote(self.provider)}")

        if params:
            self.ws_url += "?" + "&".join(params)

        # Set up the headers
        self.headers = {"Authorization": f"Bearer {config.api_key}"}

    def connect(self):
        """Establish WebSocket connection if not already connected"""
        if not self.websocket:
            print(f"Connecting to {self.ws_url}")
            self.websocket = websocket.WebSocket()
            # Pass the headers when connecting
            self.websocket.connect(self.ws_url, header=self._format_headers())

    def _format_headers(self):
        """Format headers for websocket-client"""
        return [f"{key}: {value}" for key, value in self.headers.items()]

    def close(self):
        """Close the WebSocket connection"""
        if self.websocket:
            self.websocket.close()
            self.websocket = None

    def detect(
        self,
        image: str,  # can be a URL or a base64 encoded image
        languages: List[str] = ["en"],  # e.g. ['en'], ['ch_sim', 'en']
        gpu: bool = True,
        detail: bool = True,  # True returns bounding boxes, False returns just text
        paragraph: bool = False,  # Merge text into paragraphs
        min_confidence: Optional[float] = 0.0,
    ) -> OCRResponse:
        """Synchronous method to perform OCR detection"""
        self.connect()

        if isinstance(image, Image.Image):
            buffered = BytesIO()
            image.save(buffered, format="PNG")
            image = "data:image/png;base64," + base64.b64encode(
                buffered.getvalue()
            ).decode("utf-8")

        request = OCRRequest(
            image=image,
            languages=languages,
            gpu=gpu,
            detail=detail,
            paragraph=paragraph,
            min_confidence=min_confidence,
            model=self.model,
            provider=self.provider,
        )

        if not self.websocket:
            raise ConnectionError("WebSocket connection not established")

        # Send the request
        self.websocket.send(request.model_dump_json())

        # Receive and process responses
        try:
            while True:
                response = self.websocket.recv()
                response_data = json.loads(response)

                if response_data.get("type") == OCRResponse.__name__:
                    return OCRResponse.model_validate(response_data)

                elif response_data.get("type") == ModelReadyResponse.__name__:
                    ready = ModelReadyResponse.model_validate(response_data)
                    if not ready.ready:
                        raise ValueError(f"Model not ready: {ready.error}")

                elif response_data.get("type") == ErrorResponse.__name__:
                    error = ErrorResponse.model_validate(response_data)
                    raise ValueError(f"Error: {error.error}")
                else:
                    raise ValueError(f"Unknown response type: {response_data}")

        except websocket.WebSocketException as e:
            self.close()
            raise ConnectionError("WebSocket connection closed unexpectedly") from e



from typing import Dict, List, Optional

import requests

from orign.config import GlobalConfig
from orign.models import (
    V1LlamaFactoryParams,
    V1MSSwiftBufferParams,
    V1ReplayBuffer,
    V1ReplayBufferData,
    V1ReplayBufferRequest,
    V1ReplayBuffersResponse,
    V1UpdateReplayBufferRequest,
)


class ReplayBuffer:
    def __init__(
        self,
        name: str,
        vram_request: Optional[str] = None,
        provider: str = "ms-swift",
        cpu_request: Optional[str] = None,
        trust_remote_code: Optional[bool] = None,
        adapter: Optional[str] = None,
        train_every: int = 50,
        sample_n: int = 100,
        sample_strategy: str = "Random",
        queue: Optional[str] = None,
        ms_swift_params: Optional[V1MSSwiftBufferParams] = None,
        llama_factory_params: Optional[V1LlamaFactoryParams] = None,
        labels: Optional[Dict[str, str]] = None,
        config: Optional[GlobalConfig] = None,
    ):
        name_parts = name.split("/")
        if len(name_parts) == 2:
            namespace, name = name_parts
        else:
            namespace = None
            name = name_parts[0]

        config = config or GlobalConfig.read()
        self.api_key = config.api_key
        self.orign_host = config.server

        # Construct the WebSocket URL with query parameters
        self.buffers_url = f"{self.orign_host}/v1/buffers"

        response = requests.get(
            self.buffers_url, headers={"Authorization": f"Bearer {self.api_key}"}
        )
        response.raise_for_status()
        buffers = V1ReplayBuffersResponse.model_validate(response.json())
        self.buffer = next((b for b in buffers.buffers if b.name == name), None)

        if not self.buffer:
            if not vram_request:
                raise ValueError("vram_request is required")

            request = V1ReplayBufferRequest(
                name=name,
                namespace=namespace,
                provider=provider,
                vram_request=vram_request,
                cpu_request=cpu_request,
                trust_remote_code=trust_remote_code,
                adapter=adapter,
                ms_swift_params=ms_swift_params,
                llama_factory_params=llama_factory_params,
                labels=labels,
                train_every=train_every,
                sample_n=sample_n,
                sample_strategy=sample_strategy,
                queue=queue,
            )
            response = requests.post(
                self.buffers_url,
                json=request.model_dump(),
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            response.raise_for_status()
            self.buffer = V1ReplayBuffer.model_validate(response.json())
            print(f"Created buffer {self.buffer.name}")
        else:
            print(f"Found buffer {self.buffer.name}, updating if necessary")
            request = V1UpdateReplayBufferRequest(
                name=name,
                namespace=namespace,
                provider=provider,
                vram_request=vram_request,
                cpu_request=cpu_request,
                trust_remote_code=trust_remote_code,
                adapter=adapter,
                ms_swift_params=ms_swift_params,
                llama_factory_params=llama_factory_params,
                labels=labels,
                train_every=train_every,
                sample_n=sample_n,
                sample_strategy=sample_strategy,
                queue=queue,
            )
            response = requests.patch(
                self.buffers_url,
                json=request.model_dump(),
                headers={"Authorization": f"Bearer {self.api_key}"},
            )
            response.raise_for_status()
            print(f"Updated buffer {self.buffer.name}")

    def send(self, data: List[dict], train: Optional[bool] = None):
        if not self.buffer or not self.buffer.name:
            raise ValueError("Buffer not found")

        url = f"{self.buffers_url}/{self.buffer.namespace}/{self.buffer.name}/examples"

        request = V1ReplayBufferData(examples=data, train=train)

        response = requests.post(
            url,
            json=request.model_dump(),
            headers={"Authorization": f"Bearer {self.api_key}"},
        )
        response.raise_for_status()
        return response.json()

    def train(self):
        if not self.buffer or not self.buffer.name:
            raise ValueError("Buffer not found")

        url = f"{self.buffers_url}/{self.buffer.namespace}/{self.buffer.name}/train"
        response = requests.post(
            url, headers={"Authorization": f"Bearer {self.api_key}"}
        )
        response.raise_for_status()
        return response.json()

    @classmethod
    def get(
        cls,
        namespace: Optional[str] = None,
        name: Optional[str] = None,
        config: Optional[GlobalConfig] = None,
    ) -> List[V1ReplayBuffer]:
        config = config or GlobalConfig.read()

        # Construct the WebSocket URL with query parameters
        buffers_url = f"{config.server}/v1/buffers"

        response = requests.get(
            buffers_url, headers={"Authorization": f"Bearer {config.api_key}"}
        )
        response.raise_for_status()
        buffer_response = V1ReplayBuffersResponse.model_validate(response.json())
        buffers = buffer_response.buffers
        if name:
            buffers = [b for b in buffers if b.name == name]

        if namespace:
            buffers = [b for b in buffers if b.namespace == namespace]

        return buffers



import base64
import json
from io import BytesIO
from typing import Optional
from urllib.parse import quote

import websockets
from PIL import Image

from orign.config import GlobalConfig
from orign.models import (
    EmbeddingRequest,
    EmbeddingResponse,
    ErrorResponse,
    ModelReadyResponse,
)


class AsyncEmbeddingModel:
    def __init__(
        self,
        model: Optional[str] = None,
        provider: Optional[str] = None,
    ):
        config = GlobalConfig.read()
        self.model = model
        self.provider = provider
        self.orign_host = config.server.strip("http://").strip("https://")
        self.websocket = None

        # Construct the WebSocket URL
        self.ws_url = f"wss://{self.orign_host}/v1/embedding/stream"
        params = []
        if self.model:
            params.append(f"model={quote(self.model)}")
        if self.provider:
            params.append(f"provider={quote(self.provider)}")

        if params:
            self.ws_url += "?" + "&".join(params)

    async def connect(self):
        """Establish WebSocket connection if not already connected"""
        if not self.websocket:
            config = GlobalConfig.read()
            headers = {"Authorization": f"Bearer {config.api_key}"}
            self.websocket = await websockets.connect(
                self.ws_url, extra_headers=headers
            )

    async def close(self):
        """Close the WebSocket connection"""
        if self.websocket:
            await self.websocket.close()
            self.websocket = None

    async def embed(
        self,
        text: Optional[str] = None,
        image: Optional[str | Image.Image] = None,
    ) -> EmbeddingResponse:
        await self.connect()

        if isinstance(image, Image.Image):
            buffered = BytesIO()
            image.save(buffered, format="PNG")
            image = "data:image/png;base64," + base64.b64encode(
                buffered.getvalue()
            ).decode("utf-8")

        request = EmbeddingRequest(
            text=text, image=image, model=self.model, provider=self.provider
        )

        if not self.websocket:
            raise ConnectionError("WebSocket connection not established")

        await self.websocket.send(request.model_dump_json())

        try:
            while True:
                response = await self.websocket.recv()
                response_data = json.loads(response)

                if response_data.get("type") == EmbeddingResponse.__name__:
                    resp = EmbeddingResponse.model_validate(response_data)
                    return resp
                elif response_data.get("type") == ModelReadyResponse.__name__:
                    ready = ModelReadyResponse.model_validate(response_data)
                    if not ready.ready:
                        raise ValueError(f"Model not ready: {ready.error}")
                elif response_data.get("type") == ErrorResponse.__name__:
                    error = ErrorResponse.model_validate(response_data)
                    raise ValueError(f"Error: {error.error}")
                else:
                    raise ValueError(f"Unknown response type: {response_data}")

        except websockets.exceptions.ConnectionClosed:
            await self.close()
            raise ConnectionError("WebSocket connection closed unexpectedly")



from .adapter import Adapter
from .buffer import ReplayBuffer
from .config import GlobalConfig
from .models import (
    ChatRequest,
    ChatResponse,
    EmbeddingRequest,
    EmbeddingResponse,
    OCRRequest,
    OCRResponse,
    V1Adapters,
    V1ChatEvent,
    V1Checkpoint,
    V1Datasets,
    V1DoctrParams,
    V1EasyOCRParams,
    V1LiteLLMParams,
    V1LlamaFactoryParams,
    V1ModelDeployment,
    V1ModelDeploymentRequest,
    V1ModelFile,
    V1Models,
    V1MSSwiftBufferParams,
    V1MSSwiftParams,
    V1ReplayBuffer,
    V1ReplayBufferData,
    V1ReplayBufferRequest,
    V1ReplayBuffersResponse,
    V1ReplicasInfo,
    V1SentenceTFParams,
    V1TrainingJob,
    V1TrainingJobsResponse,
    V1TrainingRequest,
    V1VLLMParams,
)
from .stream.aio.chat import AsyncChatModel
from .stream.aio.embedding import AsyncEmbeddingModel
from .stream.aio.ocr import AsyncOCRModel
from .stream.chat import ChatModel
from .stream.embedding import EmbeddingModel
from .stream.ocr import OCRModel

__all__ = [
    "AsyncChatModel",
    "AsyncEmbeddingModel",
    "AsyncOCRModel",
    "OCRModel",
    "ChatModel",
    "EmbeddingModel",
    "ReplayBuffer",
    "V1Adapters",
    "V1Checkpoint",
    "V1Datasets",
    "V1DoctrParams",
    "V1EasyOCRParams",
    "V1LiteLLMParams",
    "V1LlamaFactoryParams",
    "V1ModelDeployment",
    "V1ModelDeploymentRequest",
    "V1ModelFile",
    "V1Models",
    "V1MSSwiftBufferParams",
    "V1MSSwiftParams",
    "V1ReplayBuffer",
    "V1ReplayBufferData",
    "V1ReplayBufferRequest",
    "V1ReplayBuffersResponse",
    "V1ReplicasInfo",
    "V1SentenceTFParams",
    "V1TrainingJob",
    "V1TrainingJobsResponse",
    "V1TrainingRequest",
    "V1VLLMParams",
    "V1ChatEvent",
    "GlobalConfig",
    "ChatRequest",
    "ChatResponse",
    "EmbeddingRequest",
    "EmbeddingResponse",
    "OCRRequest",
    "OCRResponse",
    "Adapter",
]



import base64
import json
from io import BytesIO
from typing import AsyncIterator, List, Optional
from urllib.parse import quote

import websockets
from PIL import Image

from orign.config import GlobalConfig
from orign.models import (
    ChatRequest,
    ChatResponse,
    ContentItem,
    ErrorResponse,
    ImageUrlContent,
    MessageItem,
    ModelReadyResponse,
    Prompt,
    SamplingParams,
    TokenResponse,
)


class AsyncChatModel:
    def __init__(
        self,
        model: Optional[str] = None,
        provider: Optional[str] = None,
        config: Optional[GlobalConfig] = None,
    ):
        config = config or GlobalConfig.read()
        self.model = model
        self.provider = provider
        self.orign_host = config.server.strip("http://").strip("https://")
        self.websocket = None

        # Construct the WebSocket URL with query parameters
        self.ws_url = f"wss://{self.orign_host}/v1/chat/stream"
        params = []
        if self.model:
            params.append(f"model={quote(self.model)}")
        if self.provider:
            params.append(f"provider={quote(self.provider)}")

        if params:
            self.ws_url += "?" + "&".join(params)

    async def connect(self):
        """Establish WebSocket connection if not already connected"""
        if not self.websocket:
            config = GlobalConfig.read()
            headers = {"Authorization": f"Bearer {config.api_key}"}
            print("connecting to ", self.ws_url)
            self.websocket = await websockets.connect(
                self.ws_url, extra_headers=headers
            )

    async def close(self):
        """Close the WebSocket connection"""
        if self.websocket:
            await self.websocket.close()
            self.websocket = None

    async def chat(
        self,
        msg: Optional[str] = None,
        image: Optional[str | Image.Image] = None,
        prompt: Optional[Prompt] = None,
        batch: Optional[List[Prompt]] = None,
        sampling_params: Optional[SamplingParams] = None,
        stream_tokens: bool = False,
        adapter: Optional[str] = None,
    ) -> AsyncIterator[ChatResponse | TokenResponse]:
        await self.connect()

        # Create ChatRequest object
        request = ChatRequest(
            sampling_params=sampling_params or SamplingParams(),
            stream=stream_tokens,
            prompt=prompt,
            batch=batch,
            model=self.model,
            adapter=adapter,
        )

        # If msg is provided, convert it to a Prompt
        if msg:
            content = [ContentItem(type="text", text=msg)]
            if image:
                if isinstance(image, Image.Image):
                    buffered = BytesIO()
                    image.save(buffered, format="PNG")
                    image = "data:image/png;base64," + base64.b64encode(
                        buffered.getvalue()
                    ).decode("utf-8")
                content.append(
                    ContentItem(type="image_url", image_url=ImageUrlContent(url=image))
                )
            request.prompt = Prompt(
                messages=[MessageItem(role="user", content=content)]
            )
        elif prompt:
            request.prompt = prompt
        elif batch:
            request.batch = batch

        if not self.websocket:
            raise ConnectionError("WebSocket connection not established")

        # Send the request
        await self.websocket.send(request.model_dump_json())

        # Yield responses as they arrive
        try:
            while True:
                response = await self.websocket.recv()
                response_data = json.loads(response)

                if response_data.get("type") == TokenResponse.__name__:
                    resp = TokenResponse.model_validate(response_data)
                    yield resp
                    all_done = all(
                        choice.finish_reason == "stop" for choice in resp.choices
                    )
                    if all_done:
                        print("stream complete", flush=True)
                        break
                elif response_data.get("type") == ChatResponse.__name__:
                    yield ChatResponse.model_validate(response_data)
                    break
                elif response_data.get("type") == ModelReadyResponse.__name__:
                    ready = ModelReadyResponse.model_validate(response_data)
                    if not ready.ready:
                        raise ValueError(f"Model not ready: {ready.error}")
                elif response_data.get("type") == ErrorResponse.__name__:
                    error = ErrorResponse.model_validate(response_data)
                    raise ValueError(f"Error: {error.error}")
                else:
                    raise ValueError(f"Unknown response type: {response_data}")

        except websockets.exceptions.ConnectionClosed:
            await self.close()
            raise ConnectionError("WebSocket connection closed unexpectedly")



import base64
import json
from io import BytesIO
from typing import Iterator, List, Optional, Type, Union
from urllib.parse import quote

import websocket
from PIL import Image
from pydantic import BaseModel

from orign.config import GlobalConfig
from orign.models import (
    ChatRequest,
    ChatResponse,
    ContentItem,
    ErrorResponse,
    ImageUrlContent,
    MessageItem,
    ModelReadyResponse,
    Prompt,
    SamplingParams,
    TokenResponse,
)


class ChatModel:
    def __init__(
        self,
        model: Optional[str] = None,
        provider: Optional[str] = None,
        config: Optional[GlobalConfig] = None,
    ):
        config = config or GlobalConfig.read()
        self.debug = config.debug
        self.model = model
        self.provider = provider
        self.orign_host = config.server.strip("http://").strip("https://")
        self.websocket = None

        # Construct the WebSocket URL with query parameters
        self.ws_url = f"wss://{self.orign_host}/v1/chat/stream"
        params = []
        if self.model:
            params.append(f"model={quote(self.model)}")
        if self.provider:
            params.append(f"provider={quote(self.provider)}")

        if params:
            self.ws_url += "?" + "&".join(params)

        # Set up the headers
        self.headers = {"Authorization": f"Bearer {config.api_key}"}

    def connect(self):
        """Establish WebSocket connection if not already connected."""
        if not self.websocket:
            print(f"Connecting to {self.ws_url}")
            self.websocket = websocket.WebSocket()
            # Pass the headers when connecting
            self.websocket.connect(self.ws_url, header=self._format_headers())

    def _format_headers(self):
        """Format headers for websocket-client."""
        return [f"{key}: {value}" for key, value in self.headers.items()]

    def close(self):
        """Close the WebSocket connection."""
        if self.websocket:
            self.websocket.close()
            self.websocket = None

    def chat(
        self,
        msg: Optional[str] = None,
        image: Optional[str | Image.Image] = None,
        prompt: Optional[Prompt] = None,
        batch: Optional[List[Prompt]] = None,
        sampling_params: Optional[SamplingParams] = None,
        stream_tokens: bool = False,
        adapter: Optional[str] = None,
    ) -> Union[ChatResponse, Iterator[Union[TokenResponse, ChatResponse]]]:
        """Synchronous chat method."""
        self.connect()

        # Create ChatRequest object
        request = ChatRequest(
            sampling_params=sampling_params or SamplingParams(),
            stream=stream_tokens,
            prompt=prompt,
            batch=batch,
            model=self.model,
            adapter=adapter,
        )

        # If msg is provided, convert it to a Prompt
        if msg:
            content = [ContentItem(type="text", text=msg)]
            if image:
                if isinstance(image, Image.Image):
                    buffered = BytesIO()
                    image.save(buffered, format="PNG")
                    image = "data:image/png;base64," + base64.b64encode(
                        buffered.getvalue()
                    ).decode("utf-8")
                content.append(
                    ContentItem(type="image_url", image_url=ImageUrlContent(url=image))
                )
            request.prompt = Prompt(
                messages=[MessageItem(role="user", content=content)]
            )
        elif prompt:
            request.prompt = prompt
        elif batch:
            request.batch = batch

        if not self.websocket:
            raise ConnectionError("WebSocket connection not established")

        # Send the request
        self.websocket.send(request.model_dump_json())

        # Receive and process responses
        if stream_tokens:
            return self._chat_stream()
        else:
            return self._chat_once()

    def expect(
        self,
        schema: Type[BaseModel],
        msg: Optional[str] = None,
        image: Optional[str | Image.Image] = None,
        prompt: Optional[Prompt] = None,
        sampling_params: Optional[SamplingParams] = None,
        adapter: Optional[str] = None,
    ) -> Union[ChatResponse, Iterator[Union[TokenResponse, ChatResponse]]]:
        """Synchronous chat method."""
        self.connect()

        # Create ChatRequest object
        request = ChatRequest(
            sampling_params=sampling_params or SamplingParams(),
            prompt=prompt,
            model=self.model,
            adapter=adapter,
        )

        # If msg is provided, convert it to a Prompt
        if msg:
            content = [ContentItem(type="text", text=msg)]
            if image:
                if isinstance(image, Image.Image):
                    buffered = BytesIO()
                    image.save(buffered, format="PNG")
                    image = "data:image/png;base64," + base64.b64encode(
                        buffered.getvalue()
                    ).decode("utf-8")
                content.append(
                    ContentItem(type="image_url", image_url=ImageUrlContent(url=image))
                )
            request.prompt = Prompt(
                messages=[MessageItem(role="user", content=content)]
            )
        elif prompt:
            request.prompt = prompt

        if not self.websocket:
            raise ConnectionError("WebSocket connection not established")

        return self._expect_once(request, schema)

    def _expect_once(
        self,
        request: ChatRequest,
        schema: Type[BaseModel],
        max_retries: int = 3,
    ) -> ChatResponse:
        """Send the request, parse the response using the given schema, and retry if needed."""
        attempts = 0

        while attempts < max_retries:
            # If the WebSocket connection was lost between retries, ensure it's open
            if not self.websocket:
                self.connect()
            try:
                if not self.websocket:
                    raise ConnectionError("WebSocket connection not established")

                # Send the request
                self.websocket.send(request.model_dump_json())

                # Wait for and parse the single response
                response = self._chat_once()
                if self.debug:
                    print("expect response: ", response.choices[0].text, flush=True)

                response.parsed = schema.model_validate_json(response.choices[0].text)

                return response

            except (ConnectionError, ValueError, websocket.WebSocketException) as e:
                attempts += 1
                print(f"Attempt {attempts} failed with error: {e}", flush=True)

                # If we've used all retries, re-raise
                if attempts >= max_retries:
                    raise

                # Otherwise, close and reconnect before attempting again
                self.close()
                print("Reconnecting before retry...", flush=True)

        # Should never reach here since we either return or raise
        raise RuntimeError("Unexpected error in _expect_once (retry loop).")

    def _chat_once(self) -> ChatResponse:
        """Receive a single ChatResponse."""
        try:
            while True:
                response = self.websocket.recv()  # type: ignore
                response_data = json.loads(response)

                if response_data.get("type") == ChatResponse.__name__:
                    return ChatResponse.model_validate(response_data)
                elif response_data.get("type") == TokenResponse.__name__:
                    # Consume tokens until final response
                    continue
                elif response_data.get("type") == ModelReadyResponse.__name__:
                    ready = ModelReadyResponse.model_validate(response_data)
                    if not ready.ready:
                        raise ValueError(f"Model not ready: {ready.error}")
                elif response_data.get("type") == ErrorResponse.__name__:
                    error = ErrorResponse.model_validate(response_data)
                    raise ValueError(f"Error: {error.error}")
                else:
                    raise ValueError(f"Unknown response type: {response_data}")

        except websocket.WebSocketException as e:
            self.close()
            raise ConnectionError("WebSocket connection closed unexpectedly") from e

    def _chat_stream(self) -> Iterator[Union[TokenResponse, ChatResponse]]:
        """Yield responses as they arrive."""
        try:
            while True:
                response = self.websocket.recv()  # type: ignore
                response_data = json.loads(response)

                if response_data.get("type") == TokenResponse.__name__:
                    resp = TokenResponse.model_validate(response_data)
                    yield resp
                    all_done = all(
                        choice.finish_reason == "stop" for choice in resp.choices
                    )
                    if all_done:
                        print("Stream complete", flush=True)
                        break
                elif response_data.get("type") == ChatResponse.__name__:
                    yield ChatResponse.model_validate(response_data)
                    break
                elif response_data.get("type") == ModelReadyResponse.__name__:
                    ready = ModelReadyResponse.model_validate(response_data)
                    if not ready.ready:
                        raise ValueError(f"Model not ready: {ready.error}")
                elif response_data.get("type") == ErrorResponse.__name__:
                    error = ErrorResponse.model_validate(response_data)
                    raise ValueError(f"Error: {error.error}")
                else:
                    raise ValueError(f"Unknown response type: {response_data}")

        except websocket.WebSocketException as e:
            self.close()
            raise ConnectionError("WebSocket connection closed unexpectedly") from e

    def _parse_response(
        self, response: ChatResponse, model: Optional[Type[BaseModel]]
    ) -> ChatResponse:
        """Parse the response into a specific type."""
        if not model:
            return response

        # How do we handle multiple choices?
        response.parsed = model.model_validate(response.choices[0].text)
        return response



from typing import Optional

import requests

from orign.config import GlobalConfig


class Adapter:
    @classmethod
    def get(cls, name: Optional[str] = None, config: Optional[GlobalConfig] = None):
        config = config or GlobalConfig.read()

        name_parts = name.split("/") if name else []
        if len(name_parts) == 2:
            namespace, short_name = name_parts
        else:
            namespace = None
            short_name = name_parts[0]

        # Construct the WebSocket URL with query parameters
        adapters_url = f"{config.server}/v1/adapters"

        response = requests.get(
            adapters_url, headers={"Authorization": f"Bearer {config.api_key}"}
        )
        response.raise_for_status()
        response_jdict = response.json()
        adapters = response_jdict.get("adapters", [])

        if namespace:
            adapters = [a for a in adapters if a == name]
        else:
            adapters = [a for a in adapters if a.split("/")[1] == short_name]
        return adapters
